/*************************************************************/
/*   power.h                                                 */ 
/*   programmed by Weiping Liao			             */
/*   EE department, University of California at Los Angeles  */
/*   Oct, 2003                                               */
/*   Note: search for "Power model" in sim-outorder.c        */
/*************************************************************/

#ifndef POWER_H
#define POWER_H

/****** FLAG and MACRO definitions *******/

//#define _DEBUG_
//#define TMPR_DEP
#define _HOTSPOT_ // temperature calculation from hotspot
#define _CLOCKGATING_
#define _INDIVIDUAL_

#define TOTAL_MODULES 18
#define MODULE_DECODE 0
#define MODULE_BRANCH 1
#define MODULE_RAT 2
#define MODULE_RUU 3
#define MODULE_LSQ 4
#define MODULE_IALU1 5
#define MODULE_IALU2 6
#define MODULE_IALU3 7
#define MODULE_IALU4 8
#define MODULE_FALU1 9
#define MODULE_FALU2 10
#define MODULE_INTREG 11
#define MODULE_FPREG 12
#define MODULE_ITLB 13
#define MODULE_IL1 14
#define MODULE_DTLB 15
#define MODULE_DL1 16
#define MODULE_L2 17

#define MODULE_L2_LEFT MODULE_L2
#define MODULE_L2_BOTTOM MODULE_L2 + 1
#define MODULE_L2_RIGHT MODULE_L2 + 2

#define MODULE_L2_EXTRA 2 // 2 extra because L2 include 3 parts: left, right and bottom
/*#define TOTAL_COUNTERS 22
#define COUNTER_FETCH 0
#define COUNTER_DECODE 1
#define COUNTER_BRANCH 2
#define COUNTER_RAT 3
#define COUNTER_RUU_REG 4
#define COUNTER_RUU_SEL 5
#define COUNTER_RUU_WAKE 6
#define COUNTER_LSQ_REG 7
#define COUNTER_LSQ_WAKE 8
#define COUNTER_IALU1 9
#define COUNTER_IALU2 10
#define COUNTER_IALU3 11
#define COUNTER_IALU4 12
#define COUNTER_FALU1 13
#define COUNTER_FALU2 14
#define COUNTER_INTREG 15
#define COUNTER_FPREG 16
#define COUNTER_ITLB 17
#define COUNTER_IL1 18
#define COUNTER_DTLB 19
#define COUNTER_DL1 20
#define COUNTER_L2 21*/

/******* constant definitions ******/
#define OVERHEAD 1e-5 // 10us
#define OVERHEAD_FU 1e-6

// Feed-back control
#define SETPOINT 105 //81.8 // in C
#define GAIN 1

#define INTERVAL 10000 // time step 10K cycles
#define CONTROL_INTERVAL 1000000
#define T_INIT 35 //C
#define STD_FREQ 3e9 // 3GHz
#define STD_VDD 1.0
#define VDD 1.0

// since we only model SRAM array and omit the decoder/sen amp/output driver, etc, we need this remedy factor 
#define CACHE_POWER_FACTOR  3.0 // assume SRAM arrays consume about 1/3 of total cache power, 1/3 is obtained by comparing results from CACTI 

#define U 1e-6
#define SVdd 1.3
#define TECH 0.1

#define PATH_LENGTH 256
#define LINE_LENGTH 256 
#define TERM_LENGTH 64

#define DATA_WIDTH 64
#define ADDR_WIDTH 32

// MD_NUM_IREGS * 1 (each XOR) + #(priority encoder) + #(MUX)
// estimate #(priority encoder) and #(MUX) both as 2 * number of inputs 
#define DCL_GATE_COUNT 160
// number of arbiter cell is X/4 + X / 4^2 + ... + X / 4^N = (X - X / 4^N) / 3 where X is RUU_size and N is decide by 4^N = X 
// each arbiter cell includes 1 OR gate and 1 priority encoder (4 inputs, about 8 gates)
// X = 64, N = 3
#define RUU_SELECTION_GATE_COUNT 189
//#define DECODE_GATE_COUNT 1150000
#define IALU_GATE_COUNT 68215
#define FALU_GATE_COUNT 128369
#define DYNAMIC_SWITCH_FACTOR 0.15 // empirical switching factor as 0.15
#define C_GATE (0.75 + 0.91) * 1e-15 // gate cap as 0.91 fF
//#define C_GATE 0.91e-15 // gate cap as 0.91 fF
#define GATE_AREA 6.5e-12 // 6.5 um^2
#define SRAM_CELL_AREA 2e-12 // 2 um^2

#define MEM_DENSITY 0.8
#define DATA_DENSITY 1.0

#define CLOCK_GATING_REDUCTION 0.75 // 75% reduction

#define APPEND 0
#define ASSIGN 1

/****** structure definitions ******/

typedef struct COMPONENT 
{
  /* area for different circuit types, used in model I */
  double staticArea; 	/* static logic */
  double dynamicArea; 	/* dynamic logic */
  double clockArea; 	/* clock */
  double memoryArea; 	/* memory */
  double plaArea; 	/* PLA */

  double totalArea;     /* total area can be read in from configuration file, or the sum of area of five circuit types */
  double tauHeat;    	/* heating time constant */
  double tauCool;    	/* cooling time constant */
  double area; 		/* area for thermal issue */
  double thmResis;	/* thermal resistance */
 
  long access; // access counter
  long total_access; // total accesses 

  // for 100% dynamic power reduction by clock gating
  /* power */
  double activePower;    /* active power */
  double dynPower; 	 /* dynamic power */
  double standbyPower;   /* standby power */
  double inactivePower;   /* inactive power */
  double totalPower;	 /* cumulative(total) energy without any throttling */
  double totalDynamicPower; /* total dynamic power */
  double totalStandbyPower; /* total standby power */
  double totalPowerClock; /* cumulative(total) energy with only clock gating */
  double totalPowerGating; /* cumulative(total) energy with power gating, based on only utilization rate */

  /* temperature-related */
  double temperature; 	
  double lastTmprPower;

  /* temporoly variables */
  double lastPowerClock; /* power with only clock gating during last cycle */
  double lastPowerGating; /* power with power gating during last cycle */
  double presPowerClock; /* power with only clock gating during current cycle */
  double presPowerGating; /* power with power gating during current cycle */
  double presTotalPower; /* power without throttling during current cycle */

  // for 75% dynamic power reduction by clock gating
  /* power */
  double activePower_75;    /* active power */
  double dynPower_75; 	 /* dynamic power */
  double standbyPower_75;   /* standby power */
  double inactivePower_75;   /* inactive power */
  double totalPower_75;	 /* cumulative(total) energy without any throttling */
  double totalDynamicPower_75; /* total dynamic power */
  double totalStandbyPower_75; /* total standby power */
  double totalPowerClock_75; /* cumulative(total) energy with only clock gating */
  double totalPowerGating_75; /* cumulative(total) energy with power gating, based on only utilization rate */

  /* temporoly variables */
  double lastPowerClock_75; /* power with only clock gating during last cycle */
  double lastPowerGating_75; /* power with power gating during last cycle */
  double presPowerClock_75; /* power with only clock gating during current cycle */
  double presPowerGating_75; /* power with power gating during current cycle */
  double presTotalPower_75; /* power without throttling during current cycle */

  /* temperature-related */
  double temperature_75; 	
  double lastTmprPower_75;
} COMPONENT;


/* structures for memory array and CAM */
/* pre-calculate for each cache, tlb and btb */
typedef struct MEM_CONFIG
{
  int ndwl, ndbl, nspd; /* from CACTI */
  int ntwl, ntbl, ntspd; /* from CACTI */
  int rowsb, colsb; /* row and col for array: il1cac, dl1cac, btbcac ... */
  int trowsb, tcolsb; /* row and col for tag, not used for btb */
  int logGates; /* number of gates estimated for cache logic */
  int tagSize; /* bit size of tag */
  int assoc; /* associativity */
  double totalArea; /* total area */
} MEM_CONFIG;

typedef struct FORMULA
{
  double pd_a, pd_b; /* a and b coefficients for Pd */
  double ps_a, ps_b, ps_c, ps_d, ps_x, ps_y, ps_z; /* coefficients for Ps */
  double pi_a, pi_b, pi_c, pi_d, pi_x, pi_y, pi_z; /* coefficients for Pi */
} FORMULA;


/****** function declaration ******/

/* output parameters to an output file */
void output_file(char *file);

double logtwo(double x);

/* calculate standby power and update active power every cycle */
void update_standby_power(int n);

/* temperature calculation */
void output_temperature(FILE *fp);

/* power formulas */
double dyn_power_array(int row, int col, double vdd);
//double active_power_comb(int width);
double std_power_array(int row, int col, double tmpr, double vdd);
double dynamic_power_comb(int gate_count, double sf, double c_gate, double vdd);
double standby_power_comb(int gate_count, double tmpr, double vdd);
double inact_power_array(int row, int col, double tmpr, double vdd);
//double inactive_power_comb(int width, double tmpr);
void non_memory_power(COMPONENT *m, unsigned int gate_count);

void update_Tmpr(COMPONENT *m);
void update_Tmpr_75(COMPONENT *m);
void calculate_module_power();
int update_power();
void update_module_power();
void collected_power_output(FILE *fp);

void power_init();
void power_output(FILE *fp);
void output_temperature(FILE *fp);
void output_power_temperature(tick_t sim_cycle, FILE *fp);

/* tools */
int getBlock(char *name);
void reset_counter();
void power_clear(COMPONENT *c);
void update_whole_chip_tmpr(double total_power);
void update_whole_chip_tmpr_75(double total_power);
double rents(double gate_count);

/* feed-back control */
void update_dvs_dfs();
#endif
